#ifndef COCO_STREAM_LOGGER_H_
#define COCO_STREAM_LOGGER_H_

// MIT License
//
// Copyright (c) 2019 Cocotec Limited
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in all
// copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
// SOFTWARE.
//
// SPDX-License-Identifier: MIT

#include <iostream>
#include <mutex>

#include "coco/runtime.h"

namespace coco {

enum class TimestampFormat : std::uint8_t {
  None,
  ISO8601UTC,
};

/// An StateMachineLogger::EventLogger that is implemented in terms of a `Formatter` and outputs events in the Coco
/// standardised logging format.
///
/// Should be useful as a starting point for custom implementations of `EventLogger`.
class FormatterEventLogger : public StateMachineLogger::EventLogger {
 public:
  FormatterEventLogger(Formatter& formatter, TimestampFormat timestamp_format)
      : formatter_(formatter),
        timestamp_format_(timestamp_format),
        position_(Position::FirstParameter),
        port_(NULL),
        event_(NULL) {}

  virtual FormatterEventLogger& setup(const char* direction, const BaseComponent& component, const char* port,
                                      const char* event) {
    position_ = Position::FirstParameter;
    port_ = port;
    event_ = event;
    maybe_write_timestamp();
    formatter_.write_string(direction);
    formatter_.write_string(component.name());
    return *this;
  }

  void done() COCO_OVERRIDE {
    output_prefix(false);
    if (event_ && position_ != Position::Finished) {
      formatter_.write_char('}');
    }
  }

  void maybe_write_timestamp() {
    switch (timestamp_format_) {
      case TimestampFormat::None:
        break;
      case TimestampFormat::ISO8601UTC:
        formatter_.write_char('[');
        format_iso8601_utc_timestamp(std::chrono::system_clock::now(), formatter_);
        formatter_.write_string("] ");
        break;
    }
  }

  /// Prints an ISO8601-formatted timestamp to \p formatter
  static void format_iso8601_utc_timestamp(std::chrono::system_clock::time_point time_point, Formatter& formatter) {
    std::time_t native_time = std::chrono::system_clock::to_time_t(time_point);
    std::tm tm{};
    gmtime(native_time, tm);

    formatter.write_int(tm.tm_year + 1900);
    formatter.write_char('-');
    formatter.write_int_padded(tm.tm_mon + 1, 2);
    formatter.write_char('-');
    formatter.write_int_padded(tm.tm_mday, 2);
    formatter.write_char('T');
    formatter.write_int_padded(tm.tm_hour, 2);
    formatter.write_char(':');
    formatter.write_int_padded(tm.tm_min, 2);
    formatter.write_char(':');
    formatter.write_int_padded(tm.tm_sec, 2);
    {
      auto epoch_time = time_point.time_since_epoch();
      auto micros = std::chrono::duration_cast<std::chrono::microseconds>(epoch_time) -
                    std::chrono::duration_cast<std::chrono::microseconds>(
                        std::chrono::duration_cast<std::chrono::seconds>(epoch_time));

      formatter.write_char('.');
      formatter.write_int_padded(static_cast<int>(micros.count()), 6);
    }
    formatter.write_char('Z');
  }

 private:
  enum class Position : std::uint8_t {
    FirstParameter,
    NthParameter,
    Finished,
  };

  Formatter& formatter_;
  const TimestampFormat timestamp_format_;
  Position position_;
  const char* port_;
  const char* event_;

  Formatter& formatter() COCO_OVERRIDE { return formatter_; }

  void begin_parameter(const char* parameter_name) COCO_OVERRIDE {
    output_prefix(true);
    formatter_.write_string(parameter_name);
    formatter_.write_char('=');
  }
  void end_parameter(const char*) COCO_OVERRIDE {}

  void begin_return_value() COCO_OVERRIDE {
    output_prefix(false);
    formatter_.write_string("} -> ");
  }
  void end_return_value() COCO_OVERRIDE { position_ = Position::Finished; }

  void begin_current_state() COCO_OVERRIDE { formatter_.write_string(" ("); }
  void end_current_state() COCO_OVERRIDE { formatter_.write_char(')'); }

  void message(const char* message, size_t message_size) COCO_OVERRIDE {
    output_prefix(false);
    formatter().write_data(message, message_size);
  }

  void output_prefix(bool output_comma) {
    if (position_ == Position::FirstParameter) {
      formatter_.write_string(": ");
      if (port_ && port_[0] != '\0') {
        formatter_.write_string(port_);
        formatter_.write_char('.');
      }
      if (event_) {
        formatter_.write_string(event_);
        formatter_.write_char('{');
      }
      position_ = Position::NthParameter;
    } else if (output_comma) {
      formatter_.write_char(',');
    }
  }

  static inline void gmtime(const std::time_t& time, std::tm& tm) {
#ifdef _WIN32
    ::gmtime_s(&tm, &time);
#else
    ::gmtime_r(&time, &tm);
#endif
  }
};

/// A logger that outputs to a coco::Formatter in the Coco standard logging format.
class FormatterMachineLogger : public StateMachineLogger {
 public:
  void machine_starting(const BaseComponent& component) COCO_OVERRIDE { message(component, NULL, "starting machine"); }
  void machine_started(const BaseComponent& component, const char* current_state) COCO_OVERRIDE {
    message(component, current_state, "started machine");
  }
  void machine_terminated(const BaseComponent& component) COCO_OVERRIDE {
    message(component, NULL, "terminated machine");
  }
  void state_entry_started(const BaseComponent& component, const char* current_state) COCO_OVERRIDE {
    message(component, current_state, "entry started");
  }
  void state_exit_finished(const BaseComponent& component, const char* current_state) COCO_OVERRIDE {
    message(component, current_state, "exit finished");
  }
  void event_started(const BaseComponent& component, const char* current_state, const char* port,
                     const char* event) COCO_OVERRIDE {
    std::lock_guard<std::mutex> lock(mutex_);
    output_prefix(component, current_state);
    formatter_.write_string("started ");
    formatter_.write_string(port);
    formatter_.write_char('.');
    formatter_.write_string(event);
    flush_message();
  }
  void event_finished(const BaseComponent& component, const char* current_state, const char* port,
                      const char* event) COCO_OVERRIDE {
    std::lock_guard<std::mutex> lock(mutex_);
    output_prefix(component, current_state);
    formatter_.write_string("finished ");
    formatter_.write_string(port);
    formatter_.write_char('.');
    formatter_.write_string(event);
    flush_message();
  }
  void message(const BaseComponent& component, const char* current_state, const char* message) COCO_OVERRIDE {
    std::lock_guard<std::mutex> lock(mutex_);
    output_prefix(component, current_state);
    formatter_.write_string(message);
    flush_message();
  }
  void abort(const BaseComponent& component, const char* current_state) COCO_OVERRIDE {
    message(component, current_state, "abort");
  }

  EventLogger& event_started(const BaseComponent& component, const char* port, const char* event) COCO_OVERRIDE {
    return event_logger_.setup("> ", component, port, event);
  }
  EventLogger& event_finished(const BaseComponent& component, const char* port, const char* event) COCO_OVERRIDE {
    return event_logger_.setup("< ", component, port, event);
  }
  EventLogger& message(const BaseComponent& component) COCO_OVERRIDE {
    return event_logger_.setup("# ", component, NULL, NULL);
  }

 protected:
  explicit FormatterMachineLogger(Formatter& formatter, TimestampFormat timestamp_format)
      : formatter_(formatter), event_logger_(*this, timestamp_format) {}

  /// Writes a single log message.
  ///
  /// The message's contents will already have been written to `formatter_`.
  virtual void flush_message() = 0;

 private:
  class LockedEventLogger : public FormatterEventLogger {
   public:
    explicit LockedEventLogger(FormatterMachineLogger& logger, TimestampFormat timestamp_format)
        : FormatterEventLogger(logger.formatter_, timestamp_format), logger_(logger) {}

    LockedEventLogger& setup(const char* direction, const BaseComponent& component, const char* port,
                             const char* event) COCO_OVERRIDE {
      logger_.mutex_.lock();
      FormatterEventLogger::setup(direction, component, port, event);
      return *this;
    }

    void done() COCO_OVERRIDE {
      FormatterEventLogger::done();
      logger_.flush_message();
      logger_.mutex_.unlock();
    }

   private:
    FormatterMachineLogger& logger_;
  };

  std::mutex mutex_;
  Formatter& formatter_;
  LockedEventLogger event_logger_;

  void output_prefix(const BaseComponent& component, const char* current_state) {
    event_logger_.maybe_write_timestamp();
    formatter_.write_string(component.name());
    if (current_state) {
      formatter_.write_string(" (");
      formatter_.write_string(current_state);
      formatter_.write_char(')');
    }
    formatter_.write_string(": ");
  }
};

/// A simple logger that outputs to a std::ostream.
///
/// This is not intended for production usage, but as an example of how a logger could be implemented.
class StreamMachineLogger : public FormatterMachineLogger, private Formatter {
 public:
  /// Returns a pointer to a shared logger that logs to stdout.
  ///
  /// This can be directly passed to BaseComponent::set_logger_recursively, for example:
  /// @code
  ///   component.set_logger_recursively(StreamMachineLogger::cout);
  /// @endcode
  static StreamMachineLogger* cout() {
    static StreamMachineLogger logger(std::cout);
    return &logger;
  }

  /// Returns a pointer to a shared logger that logs to stderr.
  ///
  /// This can be directly passed to BaseComponent::set_logger_recursively, for example:
  /// @code
  ///   component.set_logger_recursively(StreamMachineLogger::cerr);
  /// @endcode
  static StreamMachineLogger* cerr() {
    static StreamMachineLogger logger(std::cerr);
    return &logger;
  }

  explicit StreamMachineLogger(std::ostream& out, TimestampFormat timestamp_format = TimestampFormat::None)
      : FormatterMachineLogger(static_cast<Formatter&>(*this), timestamp_format), out_(out) {}

 private:
  std::ostream& out_;

  void flush_message() COCO_OVERRIDE { out_ << std::endl; }
  void write_char(const char c) COCO_OVERRIDE { out_ << c; }
  void write_data(const char* string, size_t size) COCO_OVERRIDE {
    out_.write(string, static_cast<std::streamsize>(size));
  }
};

}  // namespace coco

#endif
